
/*
    cxadc - cx2388x adc dma driver for linux 2.4.26
    version 0.2
    developed on Linux Knoppix 2.4.26 #1 SMP Sa Apr 17 19:33:42 CEST 2004 i586 GNU/Linux

    Copyright (c) 2005 Hew How Chee <how_chee@yahoo.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
To compile     : gcc -DMODULE -D__KERNEL__ -O2 -c cxadc.c
Output file    : cxadc.o	
Create node    : mknod /dev/cxadc c 126 0 
Install driver : insmod ./cxadc.o

Run 'cxcap' to capture raw data. See cxcap.c to info on how to compile cxcap

Reference      : btaudio driver, mc4020 driver	
	
*/

/* 'dmesg' log after insmod ./cxadc.o   (Addresses/Numbers might be different)

If it is not something like this, then there might be memory allocation error.

cxadc: no of write needed 16384 no of risc page 34
PCI: Found IRQ 9 for device 00:0a.0
IRQ routing conflict for 00:0a.0, have irq 11, want irq 9
cxadc : mem addr 0xe7000000 size 0x1000000
cxadc: risc inst buff allocated at virt 0xc1280000 phy 0x1280000 size 132 kbytes
cxadc : total size allocated by get zero page 32768 kb
cxadc: end of risc inst 0xc12a000c total size 128 kbyte
cxadc: ok here
cxadc: risc init ok
cxadc: risc sync loop addr b74008 start addr b74004
cxadc: risc jmp reloop b74008
cxadc: risc start phy addr 0xB74004
cxadc: IRQ used 11
cxadc: MEM :e7000000 MMIO :cdc5d000
cxadc: init waitq 1
cxadc: dev 0x8800 (rev 5) at 00:0a.0, cxadc: irq: 11, latency: 255, mmio: 0xe7000000
cxadc: char dev register ok

*/

/* dmesg log while capturing data 
  you should see 'Interrupt stat 0x8 Masked 0x8' . If any other value appear,
  most likely there is some problem, like underflow/overflow.
  after capturing data, IRQ is not stopped, so you will keep seeing 'interrupt' and 'wake up'
  

cxadc: open [0] private_data c2fa0000
cxadc: vm end 4215a000 vm start 4015a000  size 2000000
cxadc: vm pg off 0
cxadc: mmap private data c2fa0000
cxadc : enable interrupt
cxadc : Interrupt stat 0x8 Masked 0x8
cxadc: wake up 141d
cxadc: wait for data
cxadc : Interrupt stat 0x8 Masked 0x8
cxadc: wake up 1600
cxadc: wait for data ok
cxadc: wait for data
cxadc : Interrupt stat 0x8 Masked 0x8
cxadc: wake up 1800
cxadc: wait for data ok
cxadc: wait for data
cxadc : Interrupt stat 0x8 Masked 0x8
cxadc: wake up 1a00
cxadc: wait for data ok
cxadc: wait for data
cxadc : Interrupt stat 0x8 Masked 0x8
cxadc: wake up 1c00
cxadc: wait for data ok
cxadc: wait for data
cxadc : Interrupt stat 0x8 Masked 0x8
cxadc: wake up 1e00
cxadc: wait for data ok
cxadc: wait for data
cxadc : Interrupt stat 0x8 Masked 0x8
cxadc: wake up 0
cxadc: wait for data ok
cxadc: wait for data


*/


/*
  History

  25 sep 2005 - added support for i2c, use i2c.c to tune
              - set registers to lower gain
              - set registers so that no signal is nearer to sample value 128
              - added vmux and audsel as parms during driver loading
                (for 2nd IF hardware modification, load driver using vmux=2 )
                By default audsel=2 is to route tv tuner audio signal to
                audio out of TV card, vmux=1 use the signal from video in of tv card.
                      
		
  

*/
//#include <sys/ioctl.h>
//#include <unistd.h>
//#include <fcntl.h>
//#include <sys/soundcard.h>

#include <linux/version.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/sound.h>
#include <linux/soundcard.h>
#include <linux/slab.h>
#include <linux/kdev_t.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/fs.h>
#include <linux/devfs_fs_kernel.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,18)
# define minor(x) MINOR(x)
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,23)
# define irqreturn_t void
# define IRQ_RETVAL(foobar)
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,71)
# define strlcpy(dest,src,len) strncpy(dest,src,(len)-1)
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
# define iminor(inode) minor(inode->i_rdev)
#endif


#define cx_read(adr)         readl(ctd->mmio+((adr)>>2))
#define cx_write(dat,adr)    writel((dat),(ctd->mmio+((adr)>>2)))



#define CX_SRAM_BASE 0x180000
#define CX_VIDEO_VBI_CMDS_BASE  (CX_SRAM_BASE+0x100)



/* ------------------------------------------------------------------------- */
/* CX registers                                                              */

#define CX_DEV_CNTRL2       	0x200034 //  Device control
#define CX_VID_INT_STAT     	0x200054 //  Video interrupt status
#define CX_DMA24_PTR1       	0x30008C //  DMA Current Ptr : Ch#24
#define CX_DMA24_PTR2       	0x3000CC //  DMA Tab Ptr     : Ch#24
#define CX_DMA24_CNT1       	0x30010C //  DMA Buffer Size : Ch#24
#define CX_DMA24_CNT2       	0x30014C //  DMA Table Size  : Ch#24
#define CX_VID_DMA_CNTRL       	0x31c040 //  Video DMA control
#define CX_VBI_GP_CNT        	0x31C02C //  VBI general purpose counter
#define VID_CAPTURE_CONTROL  	0x310180
#define CX_PCI_INT_STAT		0x200044
#define CX_PCI_INT_MSK		0x200040

#define CX_GPIO                 0x350010
#define CX_GPOE			0x350014	
#define CX_PINMUX_IO            0x35c044
#define CX_GP3_IO               0x35001C

#define CX_I2C_COMP_DATACNTRL   0x368000
/* ------------------------------------------------------------------------ */

#define RISC_SYNC		 0x80000000
#define RISC_WRITE		 0x10000000
#define RISC_JUMP		 0x70000000
#define RISC_WRITERM		 0xB0000000
#define RISC_WRITECM		 0xC0000000
#define RISC_WRITECR		 0xD0000000



/* ------------------------------------------------------------------------ */

//32 Mbytes VBI DMA BUFF
#define VBI_DMA_BUFF_SIZE (1024*1024*32) 
//corresponds to 8192 DMA pages of 4k bytes
#define MAX_DMA_PAGE (VBI_DMA_BUFF_SIZE/PAGE_SIZE)



/* -------------------------------------------------------------- */

#define WRITE_INST_PER_PAGE 510 
// WRITE_INST_PER_PAGE -  	1 sync instruction (first page only) followed by 510 write instruction and finally a
//  		       		1 jmp instruction to next page/reloop  
// 				this will give 4 + 510*8 +  8 = 4092 bytes, which fits into a page

#define NUMBER_OF_WRITE_PER_PAGE 510
//number of write per page must be even
//each write risc instruction is 8 byte
//and the end , there is a jmp to next page which is 8 byte
//giving a total of 510 * 8 + 8 = 4088

#define CLUSTER_BUFFER_SIZE 2048 


#define NUMBER_OF_WRITE_NEEDED (VBI_DMA_BUFF_SIZE/CLUSTER_BUFFER_SIZE)
#define NUMBER_OF_RISC_PAGE ((NUMBER_OF_WRITE_NEEDED/NUMBER_OF_WRITE_PER_PAGE)+1+1)
//+1 page for additional one write page if NUMBER_OF_WRITE_NEEDED/NUMBER_OF_WRITE_PER_PAGE is not integer
//+1 page for first sync and jmp instruction (we used 12 bytes of 4 kbytes only)










struct risc_page {
	struct risc_page * next;
	char buffer [PAGE_SIZE-sizeof(struct risc_page*)];
};
struct cxadc {
	/* linked list */
	struct cxadc *next;

	/* device info */
	int            char_dev;
	struct pci_dev *pci;
	unsigned int   irq;
	unsigned long  mem;
	unsigned long  *mmio;

	/* locking */
	int            users;
	struct semaphore lock;

	
	unsigned int    risc_inst_buff_size;
	unsigned long	*risc_inst_virt;
	dma_addr_t    	risc_inst_phy;
	
	
	
	wait_queue_head_t readQ;

	
	void *pgvec[MAX_DMA_PAGE+1];
	unsigned long *pgrisc_virt[NUMBER_OF_RISC_PAGE];
	unsigned long *pgrisc_phy[NUMBER_OF_RISC_PAGE];
	struct risc_page *risc_ptr;
};
#if 0
struct cardinfo {
	char *name;
	int rate;
	int ignore;
};
#endif
static struct cxadc *cxadcs  = NULL;
static unsigned int cxcount      = 0;
static unsigned long risc_start_phy_addr=0;
/* -------------------------------------------------------------- */




#define NUMBER_OF_CLUSTER_BUFFER 8 
#define CX_SRAM_BASE 0x180000

#define CDT_BASE  		(CX_SRAM_BASE+0x1000)
#define CLUSTER_BUFFER_BASE 	(CX_SRAM_BASE+0x4000)
#define RISC_BUFFER_BASE        (CX_SRAM_BASE+0x2000)
#define RISC_INST_QUEUE		(CX_SRAM_BASE+0x800)
#define CHN24_CMDS_BASE	0x180100
#define DMA_BUFFER_SIZE (256*1024)

#define CX_VID_INT_MSTAT 0x200058
#define CX_VID_INT_STAT 0x200054
#define CX_VID_INT_MSK 0x200050
#define INTERRUPT_MASK 0x18888


static int free_risc(struct cxadc *ctd)
{
	struct risc_page *rptr,*next;
	rptr=ctd->risc_ptr;
	while(rptr!=NULL)
	{
	
		next=rptr->next;
	
		{
			struct page *pg = virt_to_page(rptr);
			
			

    			ClearPageReserved (pg);
			free_page((unsigned long)rptr);
		}
		rptr=next;
	
	}
}

static void risc_free(struct cxadc *ctd)
{
	int i;
	for(i=0;i<NUMBER_OF_RISC_PAGE;i++)
	{
		
		if(ctd->pgrisc_virt[i])
		{
	
			struct page *pg = virt_to_page(ctd->pgrisc_virt[i]);
			
			

    			ClearPageReserved (pg);
			free_page((unsigned long)ctd->pgrisc_virt[i]);
			ctd->pgrisc_virt[i]=0;
			ctd->pgrisc_phy[i]=0;
	
		}
	
	
	}
}


static int risc_alloc(struct cxadc *ctd)
{
	int i;
	
	
	for(i=0;i<NUMBER_OF_RISC_PAGE;i++)
	{
		ctd->pgrisc_virt[i]=0;
		ctd->pgrisc_phy[i]=0;
	}
	
	for(i=0;i<NUMBER_OF_RISC_PAGE;i++)
	{
		ctd->pgrisc_virt[i]=(void*)get_zeroed_page(GFP_KERNEL); 
		if(ctd->pgrisc_virt[i]!=NULL)
		{
			ctd->pgrisc_phy[i]=(void*)virt_to_bus(ctd->pgrisc_virt[i]);
			SetPageReserved(virt_to_page(ctd->pgrisc_virt[i]));
		}
		else
		{
			risc_free(ctd);
			printk(KERN_ERR"cxadc: risc alloc failed\n");
			return 0;	
			
		}
	
	}
	return 1;

	
}


static int risc_init(struct cxadc *ctd)
{
//#define NUMBER_OF_WRITE_NEEDED (VBI_DMA_BUFF_SIZE/CLUSTER_BUFFER_SIZE)	
	unsigned int total_write;
	unsigned int nwrite;
	unsigned long *pp;
	int   index;
	unsigned long dma_addr;
	int irq_flag;
	int i;
	int pg_ind;
	int cnt_flag;
	
	#define NO_IRQ1 (0<<24)
	#define CNT_NOP (0<<16)
	#define CNT_INC (1<<16)
	#define CNT_RST (3<<16)
	#define EOL_FLAG ( 1<<26)
	#define SOL_FLAG (1<<27)
	#define IRQ1 (1<<24)
	if(!risc_alloc(ctd)) return 0;

//first page : sync and jmp	
	pp=ctd->pgrisc_virt[0];
	*pp++=RISC_SYNC|CNT_RST;
	*pp++=RISC_JUMP; 
	*pp++=(unsigned long)ctd->pgrisc_phy[1];
	
	
//the rest are write instructions 
	index=1;
	nwrite=0;
	total_write=0;
	irq_flag=0;
	pg_ind=0;
	pp=ctd->pgrisc_virt[index];
	
	
	while(1)
	{
	
		irq_flag++;
		irq_flag&=0x3ff;
		
		if(total_write==(NUMBER_OF_WRITE_NEEDED-1))
		{
			printk(KERN_INFO"cxadc: ok here\n");
			cnt_flag=CNT_RST;
		}
		else
		{
			cnt_flag= ((total_write&1) ? CNT_INC:CNT_NOP);
		}
		
		*pp++=RISC_WRITE |CLUSTER_BUFFER_SIZE| EOL_FLAG| SOL_FLAG | cnt_flag | ((irq_flag==0)?IRQ1:NO_IRQ1);
		dma_addr=virt_to_bus(ctd->pgvec[pg_ind])+( (total_write&1)?CLUSTER_BUFFER_SIZE:0);
		//printk("dma addr %x pgvec %x\n",dma_addr,ctd->pgvec[i]);
		*pp++=dma_addr;
		
		if (total_write&1) 
		{
			
			pg_ind++;
			
		}
		nwrite++;
		total_write++;
		
		if(pg_ind>=MAX_DMA_PAGE&&total_write<NUMBER_OF_WRITE_NEEDED)
		{
			pg_ind--;
			printk(KERN_ERR"cxadc: ERROR1 detected %d total write\n",total_write);
			risc_free(ctd);
			return 0;
		}
		
		if(total_write>=NUMBER_OF_WRITE_NEEDED)
		{
			*pp++=RISC_JUMP; 
			*pp++=(unsigned long)ctd->pgrisc_phy[1];	
			printk(KERN_INFO"cxadc: risc init ok\n");
			return 1;
			
		}
		if(nwrite>=WRITE_INST_PER_PAGE)
		{
			index++;
			if(index>=NUMBER_OF_RISC_PAGE)
			{
				index--;
				printk(KERN_ERR"cxadc: ERROR detected\n");
				risc_free(ctd);
				return 0;
				
			}
			
			*pp++=RISC_JUMP; 
			*pp++=(unsigned long)ctd->pgrisc_phy[index];	
			pp=ctd->pgrisc_virt[index];
			nwrite=0;
		
			///index++;
		}
		
		
	
	}
	

	
}







/*
numbuf   - number of buffer
buffsize - buffer size in bytes
buffptr  - CX sram start addr for buffer e.g. 0x182000 as in example pg 2-62 of CX23880/1/2/3 datasheet
cdtptr   - CX sram start addr for CDT    e.g. 0x181000 as in example pg 2-62 of CX23880/1/2/3 datasheet
*/


static void create_cdt_table(struct cxadc *ctd,unsigned int numbuf,unsigned int buffsize,unsigned int buffptr,unsigned int cdtptr )
{
	int i;
	unsigned int pp,qq;
	pp=buffptr;
	qq=cdtptr;
	for(i=0;i<numbuf;i++)
	{
	
		//printk("cdtptr [%x]= %x\n",qq,pp);
		
		writel(pp,ctd->mmio+(qq>>2));
		qq+=4;
		
		writel(0,ctd->mmio+(qq>>2)); /*not needed but we do it anyway*/
		qq+=4;
		
		writel(0,ctd->mmio+(qq>>2)); /*not needed but we do it anyway*/
		qq+=4;
		
		writel(0,ctd->mmio+(qq>>2)); /*not needed but we do it anyway*/
		qq+=4;
		
		
		pp+=buffsize;
		
/*readl(ctd->mmio+((adr)>>2))
#define cx_write(dat,adr)    writel((dat),(ctd->mmio+((adr)>>2)))
*/		
		
		
			
	}

}
//return 0 if unsuccessful else return start of risc instruction phy addr
static unsigned long make_risc(struct cxadc *ctd)
{
	int is_first=1;
	struct risc_page *rptr;
	struct risc_page *prev_rptr;
	long bytes_left;
	unsigned long *pp;
	unsigned long *last_pp;
	unsigned long dma_addr;
	unsigned long loop_addr;
	unsigned long risc_start_phy;
	int nwrite;
	long npage;
	int irqt;
	ctd->risc_ptr=NULL;
	
	npage=0;
	irqt=0;
	last_pp=NULL;
	prev_rptr=NULL;
	
	while(1)
	
	{
		rptr=(struct risc_page *)get_zeroed_page(GFP_KERNEL);
		//printk("alloc virt %x phy %x\n",rptr,virt_to_bus(rptr));
		if(prev_rptr==NULL)
		
		{
		
		
			ctd->risc_ptr=rptr;
		
		}
		else
		{
			prev_rptr->next=rptr;
		}
		
		if (rptr==NULL) return 0;
		
		SetPageReserved(virt_to_page(rptr));
		
		bytes_left=PAGE_SIZE - sizeof(struct risc_page *);
		
		pp=(unsigned long *) (((char *)rptr)+sizeof(struct risc_page *));
		if(last_pp)
		{
			//printk("risc jmp %x to %x\n",last_pp,virt_to_bus(rptr) +(unsigned long)pp-(unsigned long)rptr);	
			*last_pp++=RISC_JUMP|(0<<24)|(0<<16); //interrupt and increment counter
			*last_pp++=virt_to_bus(rptr) +(unsigned long)pp-(unsigned long)rptr;
			//*last_pp++;
			//*last_pp++;
			
			last_pp=NULL;
		}
	//	xxx=0; //609
		nwrite=0;
		while(bytes_left>32)
		{
		
			if (is_first)
			{
			//*pp++;
				
			
				*pp++=RISC_SYNC|(3<<16);
				is_first=0;
				bytes_left-=4;
				loop_addr=virt_to_bus(rptr) +(unsigned long)pp-(unsigned long)rptr;
				
				risc_start_phy=loop_addr-4;
				printk("cxadc: risc sync loop addr %x start addr %x\n",
					virt_to_bus(rptr) +(unsigned long)pp-(unsigned long)rptr,risc_start_phy);
				continue;
				
			}
			irqt++;
			irqt&=0x1ff;
			if(!ctd->pgvec[npage]) break;
			
			*pp++=RISC_WRITE |CLUSTER_BUFFER_SIZE|(3<<26)|(0<<16);	
			dma_addr=virt_to_bus(ctd->pgvec[npage]);
			*pp++=dma_addr;
//			*pp++;
//			*pp++;

			
			if(ctd->pgvec[npage+1]!=0)
			{
			//not last page
				*pp++=RISC_WRITE |CLUSTER_BUFFER_SIZE|(((irqt==0)?1:0)<<24)|(3<<26)|(1<<16);	
				*pp++=dma_addr+CLUSTER_BUFFER_SIZE;
//			*pp++;
//			*pp++;
				
			
			}
			else
			{
			//last page
				*pp++=RISC_WRITE |CLUSTER_BUFFER_SIZE|(((irqt==0)?1:0)<<24)|(3<<26)|(3<<16);	//reset cnt to 0
				*pp++=dma_addr+CLUSTER_BUFFER_SIZE;
///				*pp++;
//				*pp++;

			
			}
			npage++;
			bytes_left-=16;
			nwrite++;
		}
		
		//printk("nwrite %d\n",nwrite);
		if(!ctd->pgvec[npage])
		{
		//last page
		
			*pp++=RISC_JUMP|(0<<24)|(0<<16); //interrupt and increment counter
			*pp++=loop_addr;	
			//*pp++;
			//*pp++;
			printk("cxadc: risc jmp reloop %x\n",loop_addr);
			return risc_start_phy;
		}
		else
		{
			prev_rptr=rptr;
			last_pp=pp;
			
		}
	}
	
	
	
	
	
	
	
	
	
	
	
	


}
		
		
		
		
static int alloc_risc_inst_buffer(struct cxadc *ctd)
{
	//unsigned long size=
	ctd->risc_inst_buff_size=(VBI_DMA_BUFF_SIZE/CLUSTER_BUFFER_SIZE)*8+PAGE_SIZE; //add 1 page for sync instruct and jump
	ctd->risc_inst_virt=pci_alloc_consistent(ctd->pci,ctd->risc_inst_buff_size,&ctd->risc_inst_phy);
	if (ctd->risc_inst_virt==NULL)
		return -ENOMEM;
	memset(ctd->risc_inst_virt,0,ctd->risc_inst_buff_size);
	
	printk("cxadc: risc inst buff allocated at virt 0x%x phy 0x%x size %d kbytes\n",
	ctd->risc_inst_virt,ctd->risc_inst_phy,ctd->risc_inst_buff_size/1024);
	
	return 0;
}

static int free_risc_inst_buffer(struct cxadc *ctd)
{
	if(ctd->risc_inst_virt!=NULL)
	pci_free_consistent(ctd->pci,ctd->risc_inst_buff_size,ctd->risc_inst_virt,ctd->risc_inst_phy);
}
static int make_risc_instructions(struct cxadc *ctd)//,unsigned long cl_size,unsigned long max_page)
{
	int i;
	int irqt;
	unsigned long loop_addr;
	unsigned long dma_addr;
	long *pp=(long *)ctd->risc_inst_virt;

	
	
	
	
	
	
		
	loop_addr=ctd->risc_inst_phy+4;
	
	*pp++=RISC_SYNC|(3<<16);
	
	irqt=0;
	for(i=0;i<MAX_DMA_PAGE;i++)
	{
	
		irqt++;	
		irqt&=0x1ff;
		*pp++=RISC_WRITE |CLUSTER_BUFFER_SIZE|(3<<26)|(0<<16);	
		
		dma_addr=virt_to_bus(ctd->pgvec[i]);
		//printk("dma addr %x pgvec %x\n",dma_addr,ctd->pgvec[i]);
		*pp++=dma_addr;
		if(i!=MAX_DMA_PAGE-1)
		{	
			
			*pp++=RISC_WRITE |CLUSTER_BUFFER_SIZE|(((irqt==0)?1:0)<<24)|(3<<26)|(1<<16);	
			*pp++=dma_addr+CLUSTER_BUFFER_SIZE;
			
		}
		else
		{
		
			*pp++=RISC_WRITE |CLUSTER_BUFFER_SIZE|(((irqt==0)?1:0)<<24)|(3<<26)|(3<<16);	//reset cnt to 0
			*pp++=dma_addr+CLUSTER_BUFFER_SIZE;
		
		}
		
	}
	
#if 0	
	//test stop FIFO and RISC
	*pp++=RISC_WRITECR|1;
	*pp++=CX_VID_DMA_CNTRL;
	*pp++=0;
	*pp++=0xffffffff;
	//test
#endif	
//1<<24 = irq , 11<<16 = cnt
	*pp++=RISC_JUMP|(0<<24)|(0<<16); //interrupt and increment counter
	*pp++=loop_addr;
	
	printk("cxadc: end of risc inst 0x%x total size %d kbyte\n",pp,((unsigned long)pp-(unsigned long)ctd->risc_inst_virt)/1024);
	return 0;

}
/* -------------------------------------------------------------- */

static int cxadc_char_open(struct inode *inode, struct file *file)
{
	int minor = iminor(inode);
	struct cxadc *ctd;

	for (ctd = cxadcs; ctd != NULL; ctd = ctd->next)
		if (ctd->char_dev == minor)
			break;
	if (NULL == ctd)
		return -ENODEV;

	//if (debug)
		printk("cxadc: open [%d] private_data %x\n",minor,ctd);
	file->private_data = ctd;
	return 0;
}

static int cxadc_char_release(struct inode *inode, struct file *file)
{
	printk("cxadc: release\n");
	return 0;
}

static int cxadc_char_ioctl(struct inode *inode, struct file *file,
			       unsigned int cmd, unsigned long arg)
{
	struct cxadc *ctd = file->private_data;
	int ret=0,val=0,i=0;
	DECLARE_WAITQUEUE(wait, current);
	if(cmd==0x12345678)
	{
	//	unsigned int i2c;
		//printk("cxadc : arg %x\n",arg);
		
	//	printk("cxadc : i2write 0x%x\n",arg);
		cx_write(arg&3,CX_I2C_COMP_DATACNTRL);
		return 0;
	}
	
	if(cmd==0x1234567F)
	{
		cx_write(0,CX_VBI_GP_CNT);
		cx_write(1,CX_PCI_INT_MSK); //enable interrupt
		printk(KERN_INFO "cxadc : enable interrupt\n");
		return 0;
	}
	
	if(cmd==0x12345677)
	{
		
	//	printk(KERN_INFO "cxadc: wait for data\n");
	
		if (file->f_flags & O_NONBLOCK) 
		{
			return  -EAGAIN;
	
		}	
		add_wait_queue(&ctd->readQ, &wait);		
		current->state = TASK_INTERRUPTIBLE;
		schedule();
			
		if(signal_pending(current)) 
		{
			ret=  -EINTR;
				
			
		}
		
		remove_wait_queue(&ctd->readQ, &wait);
		current->state = TASK_RUNNING;	
	
		val=cx_read(CX_VBI_GP_CNT);
		if (put_user(val, (int *)arg))
		{
			return -EFAULT;
		}
		
	//	printk(KERN_INFO "cxadc: wait for data ok\n");
		
		return ret;	
		
		
		//return 0;
		//add_wait_queue(&ctd->readQ, &wait);
	}
	return 0;
}

static struct page *
cx2388x_vma_nopage (struct vm_area_struct *vma,
		   unsigned long address, int write_access)
{

	struct cxadc* ctd;
	struct page *page;
	unsigned long offset;
	offset = (address - vma->vm_start) + vma->vm_pgoff * PAGE_SIZE;
 
	ctd=vma->vm_private_data;
  	page=virt_to_page(ctd->pgvec[offset>>PAGE_SHIFT]);
  	return page;
    
	return NOPAGE_SIGBUS;
}



static struct vm_operations_struct cx2388x_vm_ops = {
	.nopage = cx2388x_vma_nopage,
};



static int 
cx2388x_mmap (struct file *file, struct vm_area_struct *vma)
{
  
  unsigned long size ; 

  
   size= (unsigned long) vma->vm_end - vma->vm_start;
   printk("cxadc: vm end %x vm start %x  size %x\n",vma->vm_end,vma->vm_start,size);
   printk("cxadc: vm pg off %x\n",vma->vm_pgoff);
    
  
  vma->vm_ops = &cx2388x_vm_ops;
  vma->vm_flags |= VM_RESERVED;
  vma->vm_private_data = file->private_data;
  printk("cxadc: mmap private data %x\n",file->private_data);
  
  return 0;
}

static struct file_operations cxadc_char_fops = {
	.owner    = THIS_MODULE,
	.llseek   = no_llseek,
	.open     = cxadc_char_open,
	.release  = cxadc_char_release,
	.ioctl    = cxadc_char_ioctl,
	.mmap     = cx2388x_mmap,
};


static char *irq_name[]={
"Y_RISC_I1","U_RISC_I1","V_RISC_I1","VBI_RISC_I1","Y_RISC_I2","U_RISC_I2","V_RISC_I2","VBI_RISC_I2",
"YF_OF","UF_OF","VF_OF","VBIF_OF","Y_SYNC","U_SYNC","V_SYNC","VBI_SYNC",
"OPC_ERR","PAR_ERR","RIP_ERR","PCI_ABORT"};
static irqreturn_t cxadc_irq(int irq, void *dev_id, struct pt_regs * regs)
{
	int count = 0;
	u32 stat,astat,ostat,allstat;
	struct cxadc *ctd = dev_id;
	int handled = 0;

	
	allstat=cx_read(CX_VID_INT_STAT);
	stat=cx_read(CX_VID_INT_MSK);
	astat=stat & allstat;
	ostat=astat;
	if(ostat!=8)
	{
		printk(KERN_INFO "cxadc : Interrupt stat 0x%x Masked 0x%x\n",allstat,ostat);
	}
	//printk("IRQ\n");	
	//allstat=cx_read(CX_VID_INT_STAT);
	if(!astat) return IRQ_RETVAL(handled); //if no interrupt bit set we return
	handled=1;
	for(count=0;count<20;count++)
	{
		
		if(astat&1)
		{
			if(count==3)
			{
				unsigned int uu=cx_read(0x310220);
//				printk("%8.8x %x %x %x %x\n",
//			uu,uu>>16,(uu>>16)&0x1f,(uu>>8)&0xff,uu&0xff);
//			printk("gadj1 %x stip3 %x gadj3 %x gadj4 %x stat %x wc %x\n ",cx_read(0x310214),cx_read(0x310210),cx_read(0x31021c),cx_read(0x310220),cx_read(0x310100),cx_read(0x31011c));
//				printk("cxadc: wake up %x\n",cx_read(CX_VBI_GP_CNT));
//				
				wake_up_interruptible(&ctd->readQ);
			}
			//printk("IRQ %s\n",irq_name[count]);
		}
		astat>>=1;
	}
	cx_write(ostat,CX_VID_INT_STAT);
	/*if(allstat&(~ostat))
	{
		printk("IRQ additional %x\n",allstat);
	}*/
	
	return IRQ_RETVAL(handled);
	
}

/* -------------------------------------------------------------- */

#define CXADC_MAX 1

static int latency = -1;
static int audsel=2;
static int vmux=1;
static int __devinit cxadc_probe(struct pci_dev *pci_dev,
				   const struct pci_device_id *pci_id)
{
	int risc_res;
	u32 regval,regaddr,i,j;
	char buffer[256];
	struct cxadc *ctd;
	void *testmalloc;
	unsigned char revision,lat;
	int rc;
	unsigned long total_size;
	int startaddr,pcilen;
	unsigned int pgsize;
	
printk("cxadc: no of write needed %d no of risc page %d\n",NUMBER_OF_WRITE_NEEDED,NUMBER_OF_RISC_PAGE);
	
	if( PAGE_SIZE!=4096)
	{
		printk(KERN_ERR "cxadc: only page size of 4096 is supported\n");
		return -EIO;

	}
	if (CXADC_MAX == cxcount)
	{
		printk(KERN_ERR "cxadc: only 1 card is supported\n");
		return -EBUSY;
	}
	if (pci_enable_device(pci_dev))
	{
		printk(KERN_ERR "cxadc: enable device failed\n");
		return -EIO;
	}
	
	if (!request_mem_region(startaddr=pci_resource_start(pci_dev,0),
				pcilen=pci_resource_len(pci_dev,0),
				"cxadc")) {
		printk(KERN_ERR "cxadc: request memory region failed\n");
		return -EBUSY;
	}
	printk(KERN_INFO "cxadc : mem addr 0x%x size 0x%x\n",startaddr,pcilen);
	
	
	ctd = kmalloc(sizeof(*ctd),GFP_ATOMIC);
	if (!ctd) {
		rc = -ENOMEM;
		printk(KERN_ERR "cxadc : kmalloc failed\n");
		goto fail0;
	}
	memset(ctd,0,sizeof(*ctd));

	ctd->pci  = pci_dev;
	ctd->irq  = pci_dev->irq;
	
	
	if(alloc_risc_inst_buffer(ctd)!=0)
	{
		printk(KERN_ERR "cxadc : cannot alloc risc buffer\n");
		rc=-ENOMEM;
		goto fail1;
	}	
	
#if 1	
	
	for(i=0;i<(MAX_DMA_PAGE+1);i++)
	{
		ctd->pgvec[i]=0;
	}
	
	total_size=0;
	pgsize=4096;
	
	for(i=0;i<MAX_DMA_PAGE;i++)
	{
		ctd->pgvec[i]=(void*)get_zeroed_page(GFP_KERNEL);
		if(ctd->pgvec[i]!=0)
		{
			int j;
			SetPageReserved (virt_to_page (ctd->pgvec[i]));
/*	 
			 for(j=0;j<20;j++)
			 {
			 	*(char *)((ctd->pgvec[i])+j)=(0x48+j/4);
			 }*/
			 //printk("allocated %x\n",pgvec[i]);
			 total_size+=pgsize;
		}
		
	}
	
	printk(KERN_INFO "cxadc : total size allocated by get zero page %d kb\n",total_size/1024);
#endif	
	make_risc_instructions(ctd);
	risc_res=risc_init(ctd);
	 	
	
	
	risc_start_phy_addr=make_risc(ctd); //not working, need to make it work
	printk("cxadc: risc start phy addr 0x%X\n",risc_start_phy_addr); 
	
	
	
	//ctd->pci  = pci_dev;
	//ctd->irq  = pci_dev->irq;
	printk("cxadc: IRQ used %d\n",ctd->irq);
	ctd->mem  = pci_resource_start(pci_dev,0);
	
	ctd->mmio = ioremap(pci_resource_start(pci_dev,0),
			    pci_resource_len(pci_dev,0));
	printk("cxadc: MEM :%x MMIO :%x\n",ctd->mem,ctd->mmio);
	//ctd->source     = 1;
	//ctd->bits       = 8;
	//ctd->channels   = 1;
	//printk("btprobe3\n");
	/* sample rate */
	//ctd->rate = card->rate;
	//if (rate[cxcount])
	//	ctd->rate = rate[cxcount];
	
	init_MUTEX(&ctd->lock);
        //init_waitqueue_head(&ctd->readq);
	init_waitqueue_head(&ctd->readQ);
	printk("cxadc: init waitq %x\n",ctd->readQ);
	if (-1 != latency) {
		printk(KERN_INFO "cxadc: setting pci latency timer to %d\n",
		       latency);
		pci_write_config_byte(pci_dev, PCI_LATENCY_TIMER, latency);
	}
	else
	{
		pci_write_config_byte(pci_dev, PCI_LATENCY_TIMER, 255);
	}
        pci_read_config_byte(pci_dev, PCI_CLASS_REVISION, &revision);
        pci_read_config_byte(pci_dev, PCI_LATENCY_TIMER, &lat);
        printk("cxadc: dev 0x%X (rev %d) at %02x:%02x.%x, ",
	       pci_dev->device,revision,pci_dev->bus->number,
	       PCI_SLOT(pci_dev->devfn),PCI_FUNC(pci_dev->devfn));
        printk("cxadc: irq: %d, latency: %d, mmio: 0x%lx\n",
	       ctd->irq, lat, ctd->mem);
	//printk("cxadc: using card config \"%s\"\n", card->name);

	/* init hw */


	pci_set_master(pci_dev);

	
	


/* we use 16kbytes of FIFO buffer*/
	create_cdt_table( ctd, NUMBER_OF_CLUSTER_BUFFER,CLUSTER_BUFFER_SIZE,CLUSTER_BUFFER_BASE,CDT_BASE);
	cx_write((CLUSTER_BUFFER_SIZE/8-1),CX_DMA24_CNT1); /* size of one buffer in qword -1 */
//	printk("cnt1:%x\n",CLUSTER_BUFFER_SIZE/8-1);
	
	
	
	cx_write(CDT_BASE,CX_DMA24_PTR2); /* ptr to cdt */
	cx_write(2*NUMBER_OF_CLUSTER_BUFFER,CX_DMA24_CNT2); /* size of cdt in qword */
	
	
	//i=cx_read(CX_DMA24_CNT1);
	//printk("dma24_cnt1 :%x\n",i);		
	//i=cx_read(CX_DMA24_PTR2);
	//printk("dma24_ptr2 :%x\n",i);		
	//i=cx_read(CX_DMA24_CNT2);
	//printk("dma24_cnt2 :%x\n",i);		
	
	
//	if(ctd->tbuf!=NULL)
	{

		unsigned int xxx;
		xxx=cx_read(CX_VID_INT_STAT);
		cx_write(xxx,CX_VID_INT_STAT); //clear interrupt	
		
#define USE_PC_MEM	
		
//		cx_write((1<<6),0x310180);
//		printk("RISC started\n");

		//memset(ctd->tbuf,0x11,16);
		//create_risc_instructions(ctd,RISC_BUFFER_BASE,(int)ctd->tbuf_dma);
#ifdef USE_PC_MEM
		
		//cx_write(risc_start_phy_addr,CHN24_CMDS_BASE); //not working but should be this one
//		if(!risc_res)
//		{
		cx_write(ctd->risc_inst_phy,CHN24_CMDS_BASE);//working 
//		}
//		else
//		{
//		cx_write((int)ctd->pgrisc_phy[0],CHN24_CMDS_BASE); //try a new risc code
//		}
#else


		cx_write(RISC_BUFFER_BASE,CHN24_CMDS_BASE);
#endif		
		cx_write(CDT_BASE,CHN24_CMDS_BASE+4);
		cx_write(2*NUMBER_OF_CLUSTER_BUFFER,CHN24_CMDS_BASE+8);
		cx_write(RISC_INST_QUEUE,CHN24_CMDS_BASE+12);
#ifdef USE_PC_MEM
		cx_write(0x40 ,CHN24_CMDS_BASE+16);
#else		
		cx_write(0x10 | 0x80000000,CHN24_CMDS_BASE+16);
#endif	
		
		//source select (see datasheet on how to change adc source)
		//1<<14 - video in
		
		vmux&=3;//default vmux=1
		cx_write((vmux<<14)|(4<<0),0x310104); //pal-B, yadc =mux1 (<<14)
		
		
		//vbi lenght CLUSTER_BUFFER_SIZE/2  work
				
		cx_write((((CLUSTER_BUFFER_SIZE)<<17)|(2<<11)),0x310188); //no of byte transferred from peripehral to fifo
								// if fifo buffer < this, it will still transfer this no of byte
								//must be multiple of 8, if not go haywire?
		
		//raw mode & byte swap <<8 (3<<8=swap)
		cx_write( ((0xe)|(0xe<<4)|(0<<8)) , 0x310184);
		cx_write(((1<<6)|(3<<1)|(1<<0)),0x310180); //capture 8 bit raw
		//cx_write(((1<<6)|(3<<1)|(1<<5)),0x310180); //capture 16 bit raw
//		cx_write(((1<<6)),0x310180);
		//run risc
		cx_write(1<<5,CX_DEV_CNTRL2);
		//enable fifo and risc
		cx_write(((1<<7)|(1<<3)),CX_VID_DMA_CNTRL);
	
	//	for(i=0x4000;i<0x8000;i+=4)
	//	{	
	//		cx_write(0xaaaaaaaa,(CX_SRAM_BASE+i));
	//	}	
		
		
			
	}
	
	
	
	

	if ((rc = request_irq(ctd->irq, cxadc_irq, SA_SHIRQ|SA_INTERRUPT,
			      "cxadc",(void *)ctd)) < 0) {
		printk(KERN_WARNING
		       "cxadc: can't request irq (rc=%d)\n",rc);
		goto fail1;
	}

#if 1
	/* register devices */
#define CX2388XADC_MAJOR  126		       
  	if (devfs_register_chrdev (CX2388XADC_MAJOR, "cxadc", &cxadc_char_fops))
    		printk (KERN_ERR "cxadc: failed to register device\n");
	else
		printk("cxadc: char dev register ok\n");	       
#endif


	//set audio multiplexer
	
	//set vbi agc
	cx_write((0<<21)|(0<<20)|(0<<19)|(4<<16)|(0x60<<8)|(0x1c<<0),0x310204);
			
	cx_write((0<<27)|(0<<26) |(1<<25)| (0x100<<16) |(0xfff<<0),   0x310200);
	cx_write((1<<23)|(0<<22)|(0<<21)|(0x0a<<16)|(0xff<<8)|(0x0<<0),0x310220);//control gain also bit 16
	cx_write((0x1c0<<17)|(0x0<<9)|(0<<7)|(0xf<<0),0x310208);
	cx_write((0x20<<17)|(0x0<<9)|(0<<7)|(0x3f<<0),0x31020c);
//	cx_write((0x1e48<<16)|(0xe0<<8)|(0x40<<0),0x310210);
	cx_write((0x1e48<<16)|(0xff<<8)|(0x0),0x310210);
	cx_write((0xe0<<17)|(0xe<<9)|(0x0<<7)|(0x7<<0),0x310214);
	cx_write((0x28<<16)|(0x28<<8)|(0x50<<0),0x31021c);//set gain of agc but not offset
	
	//==========Pixelview PlayTVPro Ultracard specific============
	//select which output is redirected to audio output jack
	//
	cx_write(1<<25,CX_GP3_IO); //use as 24 bit GPIO/GPOE 
	cx_write(0x0b,CX_GPOE); //bit 3 is to enable 4052 , bit 0-1 4052's AB
	cx_write(audsel&3,CX_GPIO); //3=audio in to audio out
	                            //2=fm out?
				    //1=silence ?
				    //0=tuner tv audio out?
	printk("cxadc: audsel = %d\n",audsel&3);
	//=================================================
	
	//i2c sda/scl set to high and use software control
	
	cx_write(3,CX_I2C_COMP_DATACNTRL);
	//=================================================
	
	/* hook into linked list */
	ctd->next = cxadcs;
	cxadcs = ctd;
	cxcount++;

	pci_set_drvdata(pci_dev,ctd);
	cx_write(INTERRUPT_MASK,CX_VID_INT_MSK);
	//cx_write(1,CX_PCI_INT_MSK); //enable interrupt
//	printk("page size %d\n",PAGE_SIZE);
        return 0;

 fail4:
	//unregister_sound_dsp(ctd->dsp_analog);
 fail3:
	//if (digital)
	//	unregister_sound_dsp(ctd->dsp_digital);
 fail2:
        free_irq(ctd->irq,ctd);	
 fail1:
 	free_risc_inst_buffer(ctd);
	kfree(ctd);
	
 fail0:
	release_mem_region(pci_resource_start(pci_dev,0),
			   pci_resource_len(pci_dev,0));
	return rc;
}

static void __devexit cxadc_remove(struct pci_dev *pci_dev)
{
	struct cxadc *ctd = pci_get_drvdata(pci_dev);
	struct cxadc *walk;
	int i;
	/* turn off all DMA / IRQs */
		//turn off pci interrupt
		cx_write(0,CX_PCI_INT_MSK);
		//turn off interrupt
		cx_write(0,CX_VID_INT_MSK);
		cx_write(~(u32)0,CX_VID_INT_STAT);
		//disable fifo and risc
		cx_write(0,CX_VID_DMA_CNTRL);
		//disable risc
		cx_write(0,CX_DEV_CNTRL2);
		
		
	/* unregister devices */

	
	
	  /* Next, unregister ourselves with the character device driver handler */
  if ((i = devfs_unregister_chrdev(CX2388XADC_MAJOR, "cxadc")))
    printk (KERN_ERR "cxadc: unregister failed, error=%d\n", -i);
    else
    printk("cxadc: unregister ok\n");

	
	/* free resources */
	free_risc_inst_buffer(ctd);
	risc_free(ctd);
#if 1	
#if 0
	free_buffer(ctd);
#endif	
#endif	
        free_irq(ctd->irq,ctd);
	
	for(i=0;i<MAX_DMA_PAGE;i++)
	{
		if(ctd->pgvec[i]) 
		{
			struct page *pg = virt_to_page(ctd->pgvec[i]);
			//printk("allocated %x\n",ctd->pgvec[i]);
			

    			ClearPageReserved (pg);
			free_page((unsigned long)ctd->pgvec[i]);
		}
	}
	free_risc(ctd);
	
	
	release_mem_region(pci_resource_start(pci_dev,0),
			   pci_resource_len(pci_dev,0));

	/* remove from linked list */
	if (ctd == cxadcs) {
		cxadcs = NULL;
	} else {
		for (walk = cxadcs; walk->next != ctd; walk = walk->next)
			; /* if (NULL == walk->next) BUG(); */
		walk->next = ctd->next;
	}
	cxcount--;

	pci_set_drvdata(pci_dev, NULL);
	kfree(ctd);
	return;
}

/* -------------------------------------------------------------- */
/* -------------------------------------------------------------- */

static struct pci_device_id cxadc_pci_tbl[] = {
        
	{
		.vendor       = 0x14f1, //conexant
		.device       = 0x8800, 
		.subvendor    = PCI_ANY_ID,
		.subdevice    = PCI_ANY_ID,
	},
	{
		/* --- end of list --- */
	}
};

static struct pci_driver cxadc_pci_driver = {
        .name     = "cxadc",
        .id_table = cxadc_pci_tbl,
        .probe    = cxadc_probe,
        .remove   = __devexit_p(cxadc_remove),
};

static int cxadc_init_module(void)
{
	return pci_module_init(&cxadc_pci_driver);
}

static void cxadc_cleanup_module(void)
{
	pci_unregister_driver(&cxadc_pci_driver);
	return;
}

module_init(cxadc_init_module);
module_exit(cxadc_cleanup_module);
MODULE_PARM(latency, "i");
MODULE_PARM(audsel, "i");
//audsel=2 is taken from tuner stereo out ?
MODULE_PARM(vmux,"i");
//vmux=2 is taken for 2nd IF (after hardware mod)
//vmux=1 is taken from video in
MODULE_PARM_DESC(latency,"pci latency timer");

MODULE_DEVICE_TABLE(pci, cxadc_pci_tbl);
MODULE_DESCRIPTION("cx2388xx adc driver");
MODULE_AUTHOR("Hew How Chee");
MODULE_LICENSE("GPL");

