
/*
	cxcap.c  version 0.2 is to be used with cxadc.c driver to capture 
	raw data from cx2388x TV tuner cards 
	  
    Copyright (c) 2005 Hew How Chee <how_chee@yahoo.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
/*

 
 The output is stored in raw.pcm as unsigned 8 bit data
 
 To compile : gcc cxcap.c -o cxcap
 To run     : ./cxcap
 
 Make sure to read instructions in cxadc.c on creating /dev/cxadc inode

  History

  4 sep 2005 - bug fix when NPAGE set to more than 8192
             - rewrite capture loop

*/


 #include <stdlib.h>
 #include <stdio.h>
 #include <fcntl.h>
 #include <sys/mman.h>



//MAX_PAGE value should be same as MAX_DMA_PAGE in cxadc.c
#define MAX_PAGE 8192   


//NPAGE is the number of 4k block use to store capture data 
//NPAGE = 1024 will enable capture of 4Mbyte of raw data
//set higher if you have enough physical memory.
//
//#define NPAGE 16384
#define NPAGE 2048
//rawbuff is the temporary buffer used to store data retrieve from cxadc circular buffer (32 Mbyte)
//the rawdata is copied into memory first (rawbuff) since the throughtput to write to harddisk
//might not be high enough.
char  rawbuff[1024*4*NPAGE];


int  main()

{	
	//int jj;
	int fd0,i;
	

	FILE *fp1;

	char *mapptr,*buffptr;
	//unsigned long size;
	int end_page,npage;
	//int diffp;
	int start_page,sync;//MAX_PAGE-1;
	
	
	
	
	
	fd0 = open ("/dev/cxadc", O_RDWR);
 	if (fd0 < 0)
 	{	
 		printf ("open error");
 		return -1;
	}
	

//map cxadc circular buffer to user space
	mapptr=mmap(0,1024*1024*32,PROT_READ|PROT_WRITE,MAP_SHARED,fd0,0x000);
	//size = ioctl (fd0, 0x1234567f, buffer); //enable interrupt
	ioctl (fd0, 0x1234567f,0); //enable interrupt
	//printf("mapptr :%x\n",mapptr);
	///xxx=0;

	//init variable
	npage=0;
	start_page=0;
	sync=1;
	buffptr=rawbuff;
	while(1)
	{

		
		
		do
		{
			ioctl (fd0, 0x12345677, &end_page); //wait for new data
			end_page&=(MAX_PAGE-1);
			
		
		} while((end_page>0x220||end_page<0x1e0)&&sync);//wait until things stabilized and ptr is at beginning of
														//cxadc circular buffer

		if(sync) printf("\ncapturing data.\n");	
		sync=0;
	
	 
	
	 	//value&=(MAX_PAGE-1);
	 	//end_page=value;
	 
	 //printf("ok  0x%x 0x%x ",start_page,end_page);
	 
		while(start_page!=end_page)
		{
			
			//copy to user buffer
			memcpy(buffptr,mapptr+(start_page*4096),4096);
			buffptr+=4096;
			//memcpy(rawbuff+(start_page*4096),mapptr+(start_page*4096),4096);
			start_page=(start_page+1)&(MAX_PAGE-1);
			npage++;
			if(npage>=NPAGE) break;
		
		}
		if(npage>=NPAGE) break;
	 
	 
	 
	 
	 
	}
	//printf("endpage 0x%x\n",end_page);
	printf("Number of 4k page captured : %d\n",npage);

	fp1=fopen("raw.pcm","wb+");	
	if(fp1==NULL)
	{
		printf("file open error\n");
		close(fd0);
		return  -1;
	}


	 	 
	fwrite(rawbuff,sizeof(rawbuff),1,fp1);
	fclose(fp1);

	close(fd0);
	return 0;

}
